(* generated by Ott 0.21 from: lang.ott *)

Require Import Arith.
Require Import Bool.
Require Import List.


Require Import Omega ClassicalChoice String basis vgtac. 
Set Implicit Arguments. 


Definition termvar := string.
Definition typebvar := nat.
Lemma eq_typebvar: forall (x y : typebvar), {x = y} + {x <> y}.
Proof.
  decide equality; auto with ott_coq_equality arith.
Defined.
Hint Resolve eq_typebvar : ott_coq_equality.

Local Notation "'termvar'" := string.

Definition eq_termvar := string_dec.
Hint Resolve eq_termvar : ott_coq_equality.



Inductive exp : Type :=  (*r expressions *)
 | Ev (x:termvar) (*r Variable *)
 | Eu : exp
 | En (n:nat)
 | Eb (b:bool)
 | Epair (e1:exp) (e2:exp)
 | Einl (e:exp)
 | Einr (e:exp)
 | Efst (e:exp)
 | Esnd (e:exp)
 | Eif (e:exp) (e1:exp) (e2:exp) (*r if *)
 | Ele (e1:exp) (e2:exp) (*r le *)
 | Eplus (e1:exp) (e2:exp) (*r plus *)
 | Eminus (e1:exp) (e2:exp) (*r minus *)
 | Emult (e1:exp) (e2:exp) (*r mult *)
 | Einput : exp
 | Eoutput (e:exp)
 | Ecase (x:termvar) (e:exp) (e1:exp) (e2:exp) (*r case *)
 | Eroll (e:exp) (*r roll *)
 | Eunroll (e:exp) (*r unroll *)
 | Efix (f:termvar) (x:termvar) (e:exp) (*r Recursive function *)
 | Eapp (e:exp) (e':exp) (*r Function application *)
 | Eseq (e:exp) (e':exp).

Inductive val : Type :=  (*r values *)
 | Vv (x:termvar)
 | Vu : val
 | Vn (n:nat)
 | Vb (b:bool)
 | Vpair (v1:val) (v2:val)
 | Vinl (v:val)
 | Vinr (v:val)
 | Vroll (v:val)
 | Vfix (f:termvar) (x:termvar) (e:exp).

Inductive oevt : Set := 
 | OQin (n:nat)
 | OQout (n:nat).

Inductive typ : Set :=  (*r types (in locally nameless style) *)
 | Tbvar (bb:typebvar) (*r Bound variable *)
 | Tunit : typ (*r Unit *)
 | Tint : typ (*r Integers *)
 | Tboo : typ (*r Booleans *)
 | Tprod (ty1:typ) (ty2:typ) (*r products *)
 | Tsum (ty1:typ) (ty2:typ) (*r Sums *)
 | Tarrow (ty1:typ) (ty2:typ) (*r Functions *)
 | Tmu (ty:typ) (*r Recursive types *).

Inductive evt : Set := 
 | Qtau : evt
 | Qobs (oq:oevt).

Definition termenv : Set := (list (termvar * typ)).


Coercion Qobs : oevt >-> evt.

(*----------------------------------------------------------------------------*)
(** Coercions *)
(*----------------------------------------------------------------------------*)

Fixpoint exp_of_val (v : val) : exp :=
  match v with
  | Vv x        => Ev x
  | Vu          => Eu
  | Vn n        => En n
  | Vb b        => Eb b
  | Vpair v1 v2 => Epair (exp_of_val v1) (exp_of_val v2)
  | Vinl v      => Einl (exp_of_val v)
  | Vinr v      => Einr (exp_of_val v)
  | Vroll v     => Eroll (exp_of_val v)
  | Vfix f x e  => Efix f x e
  end.

Structure VoE := mkVoE { e_of : exp ; v_of : val ; VoE_pf : e_of = exp_of_val v_of }.



Inductive scont : Type :=  (*r Single-step evaluation contexts *)
 | SKif (e1:exp) (e2:exp)
 | SKpairL (e:exp)
 | SKpairR (v:val)
 | SKfst : scont
 | SKsnd : scont
 | SKinl : scont
 | SKinr : scont
 | SKleL (e:exp)
 | SKleR (v:val)
 | SKplusL (e:exp)
 | SKplusR (v:val)
 | SKminusL (e:exp)
 | SKminusR (v:val)
 | SKmultL (e:exp)
 | SKmultR (v:val)
 | SKoutput : scont
 | SKcase (x:termvar) (e1:exp) (e2:exp)
 | SKroll : scont
 | SKunroll : scont
 | SKappL (v:val)
 | SKappR (e:exp)
 | SKseq (e:exp).

Definition cont : Type := list scont.

Definition voe : Type := VoE.

(*----------------------------------------------------------------------------*)
(** Definition of a ``locally closed'' type *)
(*----------------------------------------------------------------------------*)

Fixpoint lc_typ (n : nat) (ty : typ) :=
  match ty with
  | Tbvar bb    => bb < n
  | Tunit       => True 
  | Tint        => True 
  | Tboo       => True
  | (Tprod ty1 ty2)  => lc_typ n ty1 /\ lc_typ n ty2
  | (Tsum ty1 ty2)   => lc_typ n ty1 /\ lc_typ n ty2
  | (Tarrow ty1 ty2) => lc_typ n ty1 /\ lc_typ n ty2
  | (Tmu ty)     => lc_typ (S n) ty
  end.

(*----------------------------------------------------------------------------*)
(** Substitution of bounded type variables *)
(*----------------------------------------------------------------------------*)

Fixpoint tsubst1 (x: typebvar) (tyA ty: typ) :=
  match ty with
  | Tbvar bb    =>
    match nat_compare bb x with
      | Lt => Tbvar bb
      | Eq => tyA
      | Gt => Tbvar (pred bb)
    end
  | Tunit       => Tunit 
  | Tint        => Tint 
  | Tboo       => Tboo 
  | (Tprod ty1 ty2)  => Tprod  (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tsum ty1 ty2)   => Tsum   (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tarrow ty1 ty2) => Tarrow (tsubst1 x tyA ty1) (tsubst1 x tyA ty2)
  | (Tmu ty)     => Tmu     (tsubst1 (S x) tyA ty)
  end.



(** context application *)
Definition appctx_scont_exp (scont5:scont) (exp5:exp) : exp :=
  match scont5 with
  | (SKif e1 e2) => (Eif exp5 e1 e2)
  | (SKpairL e) => (Epair exp5 e)
  | (SKpairR v) => (Epair  (exp_of_val  v )  exp5)
  | SKfst => (Efst exp5)
  | SKsnd => (Esnd exp5)
  | SKinl => (Einl exp5)
  | SKinr => (Einr exp5)
  | (SKleL e) => (Ele exp5 e)
  | (SKleR v) => (Ele  (exp_of_val  v )  exp5)
  | (SKplusL e) => (Eplus exp5 e)
  | (SKplusR v) => (Eplus  (exp_of_val  v )  exp5)
  | (SKminusL e) => (Eminus exp5 e)
  | (SKminusR v) => (Eminus  (exp_of_val  v )  exp5)
  | (SKmultL e) => (Emult exp5 e)
  | (SKmultR v) => (Emult  (exp_of_val  v )  exp5)
  | SKoutput => (Eoutput exp5)
  | (SKcase x e1 e2) => (Ecase x exp5 e1 e2)
  | SKroll => (Eroll exp5)
  | SKunroll => (Eunroll exp5)
  | (SKappL v) => (Eapp exp5  (exp_of_val  v ) )
  | (SKappR e) => (Eapp e exp5)
  | (SKseq e) => (Eseq exp5 e)
end.



(*----------------------------------------------------------------------------*)
(** Definition of app_cont *)
(*----------------------------------------------------------------------------*)

Definition app_cont (k : cont) (e : exp) : exp :=
  fold_left (fun e0 k0 => appctx_scont_exp k0 e0) k e.

Definition app_cont_r k e := fold_right appctx_scont_exp e k.

Lemma app_cont_r_reverse: forall k e,
  app_cont_r k e = app_cont (rev k) e.
Proof. 
 intros; unfold app_cont, app_cont_r.
 rewrite <-fold_left_rev_right, rev_involutive; auto.
Qed.

Lemma app_cont_reverse: forall k e,
  app_cont k e = app_cont_r (rev k) e.
Proof. 
 intros; rewrite <-(rev_involutive k) at 1; rewrite app_cont_r_reverse; auto.
Qed.



(** free variables *)
Fixpoint fv_exp (e_5:exp) : list termvar :=
  match e_5 with
  | (Ev x) => (cons x nil)
  | Eu => nil
  | (En n) => nil
  | (Eb b) => nil
  | (Epair e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Einl e) => ((fv_exp e))
  | (Einr e) => ((fv_exp e))
  | (Efst e) => ((fv_exp e))
  | (Esnd e) => ((fv_exp e))
  | (Eif e e1 e2) => (app (fv_exp e) (app (fv_exp e1) (fv_exp e2)))
  | (Ele e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Eplus e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Eminus e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (Emult e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | Einput => nil
  | (Eoutput e) => ((fv_exp e))
  | (Ecase x e e1 e2) => (app (fv_exp e) (app (list_minus eq_termvar (fv_exp e1) (cons x nil)) (list_minus eq_termvar (fv_exp e2) (cons x nil))))
  | (Eroll e) => ((fv_exp e))
  | (Eunroll e) => ((fv_exp e))
  | (Efix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
  | (Eapp e e') => (app (fv_exp e) (fv_exp e'))
  | (Eseq e e') => (app (fv_exp e) (fv_exp e'))
end.

Fixpoint fv_val (v_5:val) : list termvar :=
  match v_5 with
  | (Vv x) => nil
  | Vu => nil
  | (Vn n) => nil
  | (Vb b) => nil
  | (Vpair v1 v2) => (app (fv_val v1) (fv_val v2))
  | (Vinl v) => ((fv_val v))
  | (Vinr v) => ((fv_val v))
  | (Vroll v) => ((fv_val v))
  | (Vfix f x e) => ((list_minus eq_termvar (fv_exp e) (app (cons x nil) (cons f nil))))
end.

Definition fv_scont (SK5:scont) : list termvar :=
  match SK5 with
  | (SKif e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | (SKpairL e) => ((fv_exp e))
  | (SKpairR v) => ((fv_val v))
  | SKfst => nil
  | SKsnd => nil
  | SKinl => nil
  | SKinr => nil
  | (SKleL e) => ((fv_exp e))
  | (SKleR v) => ((fv_val v))
  | (SKplusL e) => ((fv_exp e))
  | (SKplusR v) => ((fv_val v))
  | (SKminusL e) => ((fv_exp e))
  | (SKminusR v) => ((fv_val v))
  | (SKmultL e) => ((fv_exp e))
  | (SKmultR v) => ((fv_val v))
  | SKoutput => nil
  | (SKcase x e1 e2) => (app (fv_exp e1) (fv_exp e2))
  | SKroll => nil
  | SKunroll => nil
  | (SKappL v) => ((fv_val v))
  | (SKappR e) => ((fv_exp e))
  | (SKseq e) => ((fv_exp e))
end.


(** substitutions *)
Fixpoint subst_exp (sub:list (termvar*exp)) (e__6:exp) {struct e__6} : exp :=
  match e__6 with
  | (Ev x) => (match list_assoc eq_termvar x sub with None => (Ev x) | Some e_5 => e_5 end)
  | Eu => Eu 
  | (En n) => En n
  | (Eb b) => Eb b
  | (Epair e1 e2) => Epair (subst_exp sub e1) (subst_exp sub e2)
  | (Einl e) => Einl (subst_exp sub e)
  | (Einr e) => Einr (subst_exp sub e)
  | (Efst e) => Efst (subst_exp sub e)
  | (Esnd e) => Esnd (subst_exp sub e)
  | (Eif e e1 e2) => Eif (subst_exp sub e) (subst_exp sub e1) (subst_exp sub e2)
  | (Ele e1 e2) => Ele (subst_exp sub e1) (subst_exp sub e2)
  | (Eplus e1 e2) => Eplus (subst_exp sub e1) (subst_exp sub e2)
  | (Eminus e1 e2) => Eminus (subst_exp sub e1) (subst_exp sub e2)
  | (Emult e1 e2) => Emult (subst_exp sub e1) (subst_exp sub e2)
  | Einput => Einput 
  | (Eoutput e) => Eoutput (subst_exp sub e)
  | (Ecase x e e1 e2) => Ecase x (subst_exp sub e) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e1) (subst_exp (list_minus2 eq_termvar sub (cons x nil)) e2)
  | (Eroll e) => Eroll (subst_exp sub e)
  | (Eunroll e) => Eunroll (subst_exp sub e)
  | (Efix f x e) => Efix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
  | (Eapp e e') => Eapp (subst_exp sub e) (subst_exp sub e')
  | (Eseq e e') => Eseq (subst_exp sub e) (subst_exp sub e')
end.

Fixpoint subst_val (sub:list (termvar*exp)) (v_5:val) {struct v_5} : val :=
  match v_5 with
  | (Vv x) => Vv x
  | Vu => Vu 
  | (Vn n) => Vn n
  | (Vb b) => Vb b
  | (Vpair v1 v2) => Vpair (subst_val sub v1) (subst_val sub v2)
  | (Vinl v) => Vinl (subst_val sub v)
  | (Vinr v) => Vinr (subst_val sub v)
  | (Vroll v) => Vroll (subst_val sub v)
  | (Vfix f x e) => Vfix f x (subst_exp (list_minus2 eq_termvar sub (app (cons x nil) (cons f nil))) e)
end.

Definition subst_scont (sub:list (termvar*exp)) (SK5:scont) : scont :=
  match SK5 with
  | (SKif e1 e2) => SKif (subst_exp sub e1) (subst_exp sub e2)
  | (SKpairL e) => SKpairL (subst_exp sub e)
  | (SKpairR v) => SKpairR (subst_val sub v)
  | SKfst => SKfst 
  | SKsnd => SKsnd 
  | SKinl => SKinl 
  | SKinr => SKinr 
  | (SKleL e) => SKleL (subst_exp sub e)
  | (SKleR v) => SKleR (subst_val sub v)
  | (SKplusL e) => SKplusL (subst_exp sub e)
  | (SKplusR v) => SKplusR (subst_val sub v)
  | (SKminusL e) => SKminusL (subst_exp sub e)
  | (SKminusR v) => SKminusR (subst_val sub v)
  | (SKmultL e) => SKmultL (subst_exp sub e)
  | (SKmultR v) => SKmultR (subst_val sub v)
  | SKoutput => SKoutput 
  | (SKcase x e1 e2) => SKcase x (subst_exp sub e1) (subst_exp sub e2)
  | SKroll => SKroll 
  | SKunroll => SKunroll 
  | (SKappL v) => SKappL (subst_val sub v)
  | (SKappR e) => SKappR (subst_exp sub e)
  | (SKseq e) => SKseq (subst_exp sub e)
end.



Definition fv_cont (k: cont) : list termvar :=
  fold_right (fun sk fvs => fv_exp (app_cont (sk::nil) Eu) ++ fvs) nil k.

Fixpoint ble_nat (n m : nat) {struct n} : bool :=
  match n, m with
  | O, _ => true
  | S _, O => false
  | S n1, S m1 => ble_nat n1 m1
  end.

Arguments ble_nat n m : simpl never.

Coercion exp_of_val : val >-> exp.

Hint Unfold fv_cont.
Arguments fv_cont k: simpl never.

(*----------------------------------------------------------------------------*)
(** ** Canonical structures for voe *)
(*----------------------------------------------------------------------------*)

Canonical Structure VoE_var x := @mkVoE (Ev x) (Vv x) refl_equal.
Canonical Structure VoE_unit := @mkVoE (Eu) (Vu) refl_equal.
Canonical Structure VoE_int n := @mkVoE (En n) (Vn n) refl_equal.
Canonical Structure VoE_boo b := @mkVoE (Eb b) (Vb b) refl_equal.
Program Canonical Structure VoE_pair ve1 ve2 := 
  @mkVoE (Epair (e_of ve1) (e_of ve2)) (Vpair (v_of ve1) (v_of ve2)) _.
  Next Obligation. destruct ve1, ve2; subst; simpl; auto. Qed.
Program Canonical Structure VoE_inl ve1 :=
  @mkVoE (Einl (e_of ve1)) (Vinl (v_of ve1)) _.
  Next Obligation. destruct ve1; subst; simpl; auto. Qed.
Program Canonical Structure VoE_inr ve2 :=
  @mkVoE (Einr (e_of ve2)) (Vinr (v_of ve2)) _.
  Next Obligation. destruct ve2; subst; simpl; auto. Qed.
Program Canonical Structure VoE_roll ve :=
  @mkVoE (Eroll (e_of ve)) (Vroll (v_of ve)) _.
  Next Obligation. destruct ve; subst; simpl; auto. Qed.
Canonical Structure VoE_fix f x e := @mkVoE (Efix f x e) (Vfix f x e) refl_equal.
Canonical Structure VoE_val v :=  @mkVoE (exp_of_val v) v refl_equal.

Coercion VoE_val : val >-> VoE.


(** definitions *)

(* defns OpSemJudg *)
Inductive reduce : exp -> evt -> exp -> Prop :=    (* defn reduce *)
 | Rtrue : forall (e1 e2:exp),
     reduce (Eif  (Eb true)  e1 e2) Qtau e1
 | Rfalse : forall (b:bool) (e1 e2:exp),
      (  b  = false )  ->
     reduce (Eif (Eb b) e1 e2) Qtau e2
 | Rfst : forall (ve1 ve2:voe),
     reduce (Efst (Epair  (e_of  ve1 )   (e_of  ve2 ) )) Qtau  (e_of  ve1 ) 
 | Rsnd : forall (ve1 ve2:voe),
     reduce (Esnd (Epair  (e_of  ve1 )   (e_of  ve2 ) )) Qtau  (e_of  ve2 ) 
 | Rle : forall (n1 n2:nat) (b:bool),
      (  b  = ble_nat  n1   n2  )  ->
     reduce (Ele (En n1) (En n2)) Qtau (Eb b)
 | Rplus : forall (n1 n2 n:nat),
      (  n  = plus  n1   n2  )  ->
     reduce (Eplus (En n1) (En n2)) Qtau (En n)
 | Rminus : forall (n1 n2 n:nat),
      (  n  = minus  n1   n2  )  ->
     reduce (Eminus (En n1) (En n2)) Qtau (En n)
 | Rmult : forall (n1 n2 n:nat),
      (  n  = mult  n1   n2  )  ->
     reduce (Emult (En n1) (En n2)) Qtau (En n)
 | Rinput : forall (n:nat),
     reduce Einput  (Qobs  (OQin n) )  (En n)
 | Routput : forall (n:nat),
     reduce (Eoutput (En n))  (Qobs  (OQout n) )  Eu
 | Rinl : forall (x:termvar) (ve:voe) (e1 e2:exp),
     reduce (Ecase x (Einl  (e_of  ve ) ) e1 e2) Qtau  (subst_exp(( x ,  (e_of  ve )  )::nil) e1 ) 
 | Rinr : forall (x:termvar) (ve:voe) (e1 e2:exp),
     reduce (Ecase x (Einr  (e_of  ve ) ) e1 e2) Qtau  (subst_exp(( x ,  (e_of  ve )  )::nil) e2 ) 
 | Rbeta : forall (f x:termvar) (e:exp) (ve:voe),
     reduce (Eapp  ( (Efix f x e) )   (e_of  ve ) ) Qtau  (subst_exp(( x ,  (e_of  ve )  )::( f , (Efix f x e) )::nil) e ) 
 | Rseq : forall (ve:voe) (e:exp),
     reduce (Eseq  (e_of  ve )  e) Qtau e
 | Rroll : forall (ve:voe),
     reduce (Eunroll  ( (Eroll  (e_of  ve ) ) ) ) Qtau  (e_of  ve ) 
 | Rscont : forall (SK:scont) (e:exp) (q:evt) (e':exp)
     (RED: reduce e q e'),
     reduce  (appctx_scont_exp  SK   e )  q  (appctx_scont_exp  SK   e' ) .


Hint Constructors reduce. 

Structure taggedExp := tExp0 { untagExp :> exp}.

Structure skes := mkskes
  { term_of : taggedExp; scont_of : scont; body_of : exp;
    KEsplit_pf : untagExp term_of = app_cont [scont_of] body_of }.

Structure kes := mkkes
  { exp_of : taggedExp; cont_of : cont; redex_of : exp;
    ke_split : untagExp exp_of = app_cont cont_of redex_of;
    notval : forall ve : voe, ~ e_of ve = untagExp exp_of }.

Definition tExp1  e := tExp0  e.
Definition tExp2  e := tExp1  e.
Definition tExp3  e := tExp2  e.
Definition tExp4  e := tExp3  e.
Definition tExp5  e := tExp4  e.
Definition tExp6  e := tExp5  e.
Definition tExp7  e := tExp6  e.
Definition tExp8  e := tExp7  e.
Definition tExp9  e := tExp8  e.
Definition tExp10 e := tExp9  e.
Definition tExp11 e := tExp10 e.
Definition tExp12 e := tExp11 e.
Definition tExp13 e := tExp12 e.
Definition tExp14 e := tExp13 e.
Definition tExp15 e := tExp14 e.
Definition tExp16 e := tExp15 e.
Definition tExp17 e := tExp16 e.
Definition tExp18 e := tExp17 e.
Definition tExp19 e := tExp18 e.
Definition tExp20 e := tExp19 e.
Definition tExp21 e := tExp20 e.
Definition tExp22 e := tExp21 e.
Definition tExp23 e := tExp22 e.
Definition tExp24 e := tExp23 e.
Definition tExp25 e := tExp24 e.
Definition tExp26 e := tExp25 e.
Definition tExp27 e := tExp26 e.
Definition tExp28 e := tExp27 e.
Definition tExp29 e := tExp28 e.
Definition tExp30 e := tExp29 e.
Definition tExp31 e := tExp30 e.
Definition tExp32 e := tExp31 e.
Definition tExp33 e := tExp32 e.
Definition tExp34 e := tExp33 e.
Definition tExp35 e := tExp34 e.
Canonical Structure tExp36 e := tExp35 e.

(*----------------------------------------------------------------------------*)
(** ** simplcanon  *)
(*----------------------------------------------------------------------------*)

Lemma exp_of_val_inj: forall v v', exp_of_val v = exp_of_val v' -> v = v'.
Proof.
  induction v; destruct v'; intross; clarify; try by rewrite (IHv _ H).
  by rewrite (IHv1 _ H1), (IHv2 _ H).
Qed.

Lemma e_of_inj :
  forall v v', e_of v = e_of v' -> v_of v = v_of v'.
Proof. by i; destruct v, v'; subst; apply exp_of_val_inj in H; simpls; subst. Qed.

Lemma v_of_inj :
  forall v v', v_of v = v_of v' -> v = v'.
Proof. by intros; destruct v, v'; simpls; subst. Qed.

Ltac simplcanon := 
  subst;
  (repeat match goal with 
   | [H : voe |- _] => let e := fresh "e" in let v := fresh "v" in destruct H as (e,v,?) 
   | [H : kes |- _] => let k := fresh "k" in let e := fresh "e" in 
                       destruct H as (?, k, e, ?, ?)
   | [H : skes|- _] => let sk := fresh "sk" in let e := fresh "e" in 
                       destruct H as (?, sk, e, ?)
   | [H : taggedExp |- _] => let e := fresh "e" in destruct H as (e)
   end); 
  subst; 
  repeat match goal with
  | H : v_of _ = v_of _ |- _ => 
      try apply v_of_inj in H; let P := type of H in fold (__HID__ P) in H
  | H : _ = _ |- _ => 
      try apply e_of_inj in H; let P := type of H in fold (__HID__ P) in H
  end; unfold __HID__ in *;
  simpl v_of in *; simpl e_of in *; 
  simpl exp_of in *; simpl cont_of in *; simpl redex_of in *; 
  simpl term_of in *; simpl scont_of in *; simpl body_of in *; 
  simpl untagExp in *; subst.

(*----------------------------------------------------------------------------*)
(** ** Canonical structures for skes *)
(*----------------------------------------------------------------------------*)

Canonical Structure skes_SKif ee e1 e2 :=
  mkskes (tExp0 (Eif ee e1 e2)) (SKif e1 e2) ee refl_equal.
Program 
Canonical Structure skes_SKpairR ee ve :=
  mkskes (tExp1 (Epair (e_of ve) ee)) (SKpairR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKpairL ee e :=
  mkskes (tExp2 (Epair ee e)) (SKpairL e) ee refl_equal.
Canonical Structure skes_SKfst ee :=
  mkskes (tExp3 (Efst ee)) SKfst ee refl_equal.
Canonical Structure skes_SKsnd ee :=
  mkskes (tExp4 (Esnd ee)) SKsnd ee refl_equal.
Canonical Structure skes_SKinl ee :=
  mkskes (tExp5 (Einl ee)) SKinl ee refl_equal.
Canonical Structure skes_SKinr ee :=
  mkskes (tExp6 (Einr ee)) SKinr ee refl_equal.
Program 
Canonical Structure skes_SKleR ee ve :=
  mkskes (tExp7 (Ele (e_of ve) ee)) (SKleR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKleL ee e :=
  mkskes (tExp8 (Ele ee e)) (SKleL e) ee refl_equal.
Program 
Canonical Structure skes_SKplusR ee ve :=
  mkskes (tExp9 (Eplus (e_of ve) ee)) (SKplusR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKplusL ee e :=
  mkskes (tExp10 (Eplus ee e)) (SKplusL e) ee refl_equal.
Program 
Canonical Structure skes_SKminusR ee ve :=
  mkskes (tExp11 (Eminus (e_of ve) ee)) (SKminusR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKminusL ee e :=
  mkskes (tExp12 (Eminus ee e)) (SKminusL e) ee refl_equal.
Program 
Canonical Structure skes_SKmultR ee ve :=
  mkskes (tExp13 (Emult (e_of ve) ee)) (SKmultR (v_of ve)) ee _.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKmultL ee e :=
  mkskes (tExp14 (Emult ee e)) (SKmultL e) ee refl_equal.
Canonical Structure skes_SKoutput ee :=
  mkskes (tExp15 (Eoutput ee)) SKoutput ee refl_equal.
Canonical Structure skes_SKcase ee x e1 e2 :=
  mkskes (tExp16 (Ecase x ee e1 e2)) (SKcase x e1 e2) ee refl_equal.
Canonical Structure skes_SKroll ee :=
  mkskes (tExp17 (Eroll ee)) SKroll ee refl_equal.
Canonical Structure skes_SKunroll ee :=
  mkskes (tExp18 (Eunroll ee)) SKunroll ee refl_equal.
Program 
Canonical Structure skes_SKappL ee ve :=
  mkskes (tExp19 (Eapp ee (e_of ve))) (SKappL (v_of ve)) ee refl_equal.
Next Obligation. destruct ve; simpl; subst; auto. Qed.
Canonical Structure skes_SKappR ee e :=
  mkskes (tExp20 (Eapp e ee)) (SKappR e) ee refl_equal.
Canonical Structure skes_SKseq ee e :=
  mkskes (tExp21 (Eseq ee e)) (SKseq e) ee refl_equal.
Canonical Structure skes_SK SK e :=
  mkskes (tExp22 (appctx_scont_exp SK e)) SK e refl_equal.

(*----------------------------------------------------------------------------*)
(** ** Canonical structures for kes *)
(*----------------------------------------------------------------------------*)

Program Canonical Structure kes_SKif ke e1 e2 :=
  mkkes (tExp0 (Eif (exp_of ke) e1 e2)) (cont_of ke ++ [SKif e1 e2]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKpairR ke (ve: voe) :=
  mkkes (tExp1 (Epair (e_of ve) (exp_of ke))) (cont_of ke ++ [SKpairR (v_of ve)]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKpairL ke e :=
  mkkes (tExp2 (Epair (exp_of ke) e)) (cont_of ke ++ [SKpairL e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKfst ke :=
  mkkes (tExp3 (Efst (exp_of ke))) (cont_of ke ++ [SKfst]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKsnd ke :=
  mkkes (tExp4 (Esnd (exp_of ke))) (cont_of ke ++ [SKsnd]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKinl ke :=
  mkkes (tExp5 (Einl (exp_of ke))) (cont_of ke ++ [SKinl]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKinr ke :=
  mkkes (tExp6 (Einr (exp_of ke))) (cont_of ke ++ [SKinr]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKleR ke (ve: voe) :=
  mkkes (tExp7 (Ele (e_of ve) (exp_of ke))) (cont_of ke ++ [SKleR (v_of ve)]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKleL ke e :=
  mkkes (tExp8 (Ele (exp_of ke) e)) (cont_of ke ++ [SKleL e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKplusR ke (ve: voe) :=
  mkkes (tExp9 (Eplus (e_of ve) (exp_of ke))) (cont_of ke ++ [SKplusR (v_of ve)]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKplusL ke e :=
  mkkes (tExp10 (Eplus (exp_of ke) e)) (cont_of ke ++ [SKplusL e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKminusR ke (ve: voe) :=
  mkkes (tExp11 (Eminus (e_of ve) (exp_of ke))) (cont_of ke ++ [SKminusR (v_of ve)]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKminusL ke e :=
  mkkes (tExp12 (Eminus (exp_of ke) e)) (cont_of ke ++ [SKminusL e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKmultR ke (ve: voe) :=
  mkkes (tExp13 (Emult (e_of ve) (exp_of ke))) (cont_of ke ++ [SKmultR (v_of ve)]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKmultL ke e :=
  mkkes (tExp14 (Emult (exp_of ke) e)) (cont_of ke ++ [SKmultL e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKoutput ke :=
  mkkes (tExp15 (Eoutput (exp_of ke))) (cont_of ke ++ [SKoutput]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKcase ke x e1 e2 :=
  mkkes (tExp16 (Ecase x (exp_of ke) e1 e2)) (cont_of ke ++ [SKcase x e1 e2]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKroll ke :=
  mkkes (tExp17 (Eroll (exp_of ke))) (cont_of ke ++ [SKroll]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKunroll ke :=
  mkkes (tExp18 (Eunroll (exp_of ke))) (cont_of ke ++ [SKunroll]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKappL ke (ve: voe) :=
  mkkes (tExp19 (Eapp (exp_of ke) (e_of ve))) (cont_of ke ++ [SKappL (v_of ve)]) 
       (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKappR ke e :=
  mkkes (tExp20 (Eapp e (exp_of ke))) (cont_of ke ++ [SKappR e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.
Program Canonical Structure kes_SKseq ke e :=
  mkkes (tExp21 (Eseq (exp_of ke) e)) (cont_of ke ++ [SKseq e]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v; inv FH; eauto. Qed.

Program Canonical Structure kes_SK sk ke :=
  mkkes (tExp22 (appctx_scont_exp sk (exp_of ke))) (cont_of ke ++ [sk]) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation. simplcanon; destruct v, sk; inv FH; eauto. Qed.

Program Canonical Structure kes_K k ke :=
  mkkes (tExp23 (app_cont k (exp_of ke))) (cont_of ke ++ k) (redex_of ke) _ _.
  Next Obligation. simplcanon; setoid_rewrite fold_left_app; eauto. Qed.
  Next Obligation.
    simplcanon; revert v notval0 FH; generalize (app_cont k0 e0); clear k0 e0.
    induction k; eauto.
    i; eapply IHk in FH; eauto.
    i; simplcanon; destruct v0, a; inv FH0; eauto.
  Qed.

Program Canonical Structure kes_fst e := 
  @mkkes (tExp24 (Efst e)) nil (Efst e) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_snd e := 
  @mkkes (tExp25 (Esnd e)) nil (Esnd e) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_if e e1 e2 := 
  @mkkes (tExp26 (Eif e e1 e2)) nil (Eif e e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_le e1 e2 := 
  @mkkes (tExp27 (Ele e1 e2)) nil (Ele e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_plus e1 e2 := 
  @mkkes (tExp28 (Eplus e1 e2)) nil (Eplus e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_minus e1 e2 := 
  @mkkes (tExp29 (Eminus e1 e2)) nil (Eminus e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_mult e1 e2 := 
  @mkkes (tExp30 (Emult e1 e2)) nil (Emult e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_input := 
  @mkkes (tExp31 Einput) nil Einput _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_output e := 
  @mkkes (tExp32 (Eoutput e)) nil (Eoutput e) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_case x e e1 e2 := 
  @mkkes (tExp33 (Ecase x e e1 e2)) nil (Ecase x e e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_unroll e := 
  @mkkes (tExp34 (Eunroll e)) nil (Eunroll e) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_app e1 e2 := 
  @mkkes (tExp35 (Eapp e1 e2)) nil (Eapp e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.
Program Canonical Structure kes_seq e1 e2 := 
  @mkkes (tExp36 (Eseq e1 e2)) nil (Eseq e1 e2) _ _.
  Next Obligation. simplcanon; destruct v; inv FH. Qed.

(*----------------------------------------------------------------------------*)
(** ** symbolic execution *)
(*----------------------------------------------------------------------------*)

Lemma Rcont :
  forall e q e' (RED: reduce e q e') k,
     reduce (app_cont k e) q (app_cont k e').
Proof. intros; revert e e' RED; induction k; intros; eauto using Rscont. Qed.
Hint Resolve Rcont.

Lemma reduce_scont: forall ske q e'
    (RED: reduce (body_of ske) q e'),
  reduce (term_of ske) q (appctx_scont_exp (scont_of ske) e').
Proof. i; simplcanon; s; eauto. Qed.
Hint Resolve reduce_scont.

Lemma reduce_cont: forall ke q e'
    (RED: reduce (redex_of ke) q e'),
  reduce (exp_of ke) q (app_cont (cont_of ke) e').
Proof. i; simplcanon; clear notval0. revert e q e' RED; induction k; intros; eauto. Qed.

Ltac step := by eapply reduce_cont; s; econs.

(*----------------------------------------------------------------------------*)
(** ** Notation *)
(*----------------------------------------------------------------------------*)

Notation "e1 @ e2" := ((Eapp (e1) (e2))) (at level 46).
Notation "e1 ~* e2" := ((Emult (e1) (e2))) (at level 47).
Notation "e1 ~+ e2" := ((Eplus (e1) (e2))) (at level 48).
Notation "e1 ~- e2" := ((Eminus (e1) (e2))) (at level 48).
Notation "e1 ~<= e2" := ((Ele (e1) (e2))) (at level 49).
Notation "<>" := (Eu) (at level 50).
Notation "<| e1 , e2 |>" := ((Epair (e1) (e2))) (at level 50).
Notation "'Let' x ':=' e 'in' e'" := (((Efix x x e') @ (e))) (at level 52).
Notation "'If' e 'then' e1 'else' e2" := ((Eif (e) (e1) (e2))) (at level 51).
Notation "'Case' x ':=' e '|=>' e1 '|=>' e2" := ((Ecase (x) (e) (e1) (e2))) (at level 51).
Notation "e1 ;; e2" := (Eseq (e1) (e2)) (at level 50).

Coercion En : nat >-> exp.
Coercion Eb : bool >-> exp.
Coercion Ev : string >-> exp.



