Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.
Require Import pacotac.
Require Import simul_def.

Set Implicit Arguments.

(******************************
   parameterized similarity
 ******************************)

CoInductive psimil r (e1 e2: exp) : Prop :=
| psimil_fold
  (VAL: forall (ve: voe) (VAL: e_of ve = e1), taureduces e2 (e_of ve))
  (EXP: forall q e1' (RED: reduce e1 q e1'),
        exists e2m e2', 
          <<TAU: taureduces e2 e2m>> /\ 
          <<RED: qreduce e2m q e2'>> /\
          <<REL: (psimil r \2/ r) e1' e2'>>)
.

Theorem psimil_acc: forall
    l r0 (OBG: forall r (INC: r0 <2= r) (CIH: l <2= r), l <2= psimil r),
  l <2= psimil r0.
Proof.
  intros; assert (SIM: psimil (r0 \2/ l) x0 x1) by eauto.
  clear PR; revert x0 x1 SIM; cofix CIH.
  intros; destruct SIM; econstructor; eauto.
  intros; eapply EXP in RED; des; esplits; eauto 10.
Qed.

(**********************************************
   Verification using direct internal approach
 **********************************************)

Tactic Notation "psimil_step" int_or_var(n) := 
  try left; simul_step n.

Lemma fg_simulated: forall n m (EQ: n = 2 * m), 
  psimil bot2 (f @ n) (g @ m).
Proof.
  pcofix CIH using psimil_acc.
  intros; subst; do 6 psimil_step 1; do 2 psimil_step 0.
  destruct m0; psimil_step 2; [|eauto].
  left; fold restart f g; generalize (m+(m+0)).
  pcofix CIH' using psimil_acc.
  intros; do 3 psimil_step 1.
  destruct n; psimil_step 1; [eauto|].
  do 3 psimil_step 1; eauto.
Qed.
