Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.
Require Import paconotation simul_def.

Set Implicit Arguments.

(******************************
   similarity
 ******************************)

CoInductive similarity (e1 e2: exp) : Prop :=
| sim_fold (SIM: sim similarity e1 e2)
.

Lemma simul_tarski: forall (l: exprel)
    (OBG: l <2= sim l),
  l <2= similarity.
Proof. 
  cofix CIH; i; econs; destruct (OBG _ _ PR). 
  econs; eauto; i; exploit EXP; i; des; eauto 10. 
Qed.

(********************************************
   Verification using tarski theorem (sumulation relation)
 ********************************************)

Inductive fgsim : exp -> exp -> Prop :=
| _fgsim1: forall (n m: nat) (EQ: n = 2 * m), fgsim 
     (f @ n) 
     (g @ m)
| _fgsim2: forall (m: nat), fgsim 
     (Let "v" := (Eoutput (m + (m + 0));; Einput) ~* 2
      in (If 1 ~<= "v" then f else restart @ f) @ ("v" ~+ (m + (m + 0))))
     (Eoutput (2 ~* m);;
      (Let "v" := Einput
       in If "v" ~<= 0 then (restart @ g) @ (2 ~* m) else g @ ("v" ~+ m)))
| _fgsim3: forall (m: nat), fgsim 
     (Let "v" := ((<>);; Einput) ~* 2
      in (If 1 ~<= "v" then f else restart @ f) @ ("v" ~+ (m + (m + 0))))
     ((<>);;
      (Let "v" := Einput
       in If "v" ~<= 0 then (restart @ g) @ (2 ~* m) else g @ ("v" ~+ m)))
| _fgsim4: forall (m: nat), fgsim 
     (Let "v" := Einput ~* 2
      in (If 1 ~<= "v" then f else restart @ f) @ ("v" ~+ (m + (m + 0))))
     (Let "v" := Einput
      in If "v" ~<= 0 then (restart @ g) @ (2 ~* m) else g @ ("v" ~+ m))
| _fgsim5: forall (m m0: nat), fgsim 
     (Let "v" := m0 ~* 2
      in (If 1 ~<= "v" then f else restart @ f) @ ("v" ~+ (m + (m + 0))))
     (Let "v" := m0
      in If "v" ~<= 0 then (restart @ g) @ (2 ~* m) else g @ ("v" ~+ m))
| _fgsim6: forall (m m0: nat), fgsim 
     (Let "v" := m0 * 1 + m0
      in (If 1 ~<= "v" then f else restart @ f) @ ("v" ~+ (m + (m + 0))))
     (If m0 ~<= 0 then (restart @ g) @ (2 ~* m) else g @ (m0 ~+ m))
| _fgsim7: forall (m m0: nat), fgsim 
     ((If 1 ~<= (m0 * 1 + m0) then f else restart @ f) @
      ((m0 * 1 + m0) ~+ (m + (m + 0))))
     (If ble_nat m0 0 then (restart @ g) @ (2 ~* m) else g @ (m0 ~+ m))
| _fgsim8: forall (m m0: nat), fgsim 
     ((If 1 ~<= (m0 * 1 + m0) then f else restart @ f) @
      (m0 * 1 + m0 + (m + (m + 0))))
     (If ble_nat m0 0 then (restart @ g) @ (2 ~* m) else g @ (m0 ~+ m))
| _fgsim9: forall (m m0: nat), fgsim 
     ((If ble_nat 1 (m0 * 1 + m0) then f else restart @ f) @
      (m0 * 1 + m0 + (m + (m + 0))))
     (If ble_nat m0 0 then (restart @ g) @ (2 ~* m) else g @ (m0 ~+ m))
| _fgsim10: forall (m: nat), fgsim 
     ((restart @ f) @ m) ((restart @ g) @ m)
| _fgsim11: forall (m: nat), fgsim 
     (Efix "_" "n"
        (If "n" ~<= 0 then f @ 0
         else Eoutput "n";; (restart @ f) @ ("n" ~- 1)) @ m)
     (Efix "_" "n"
        (If "n" ~<= 0 then g @ 0
         else Eoutput "n";; (restart @ g) @ ("n" ~- 1)) @ m)
| _fgsim12: forall (m: nat), fgsim 
     (If m ~<= 0 then f @ 0 else Eoutput m;; (restart @ f) @ (m ~- 1))
     (If m ~<= 0 then g @ 0 else Eoutput m;; (restart @ g) @ (m ~- 1))
| _fgsim13: forall (m: nat), fgsim 
     (If ble_nat m 0 then f @ 0 else Eoutput m;; (restart @ f) @ (m ~- 1))
     (If ble_nat m 0 then g @ 0 else Eoutput m;; (restart @ g) @ (m ~- 1))
| _fgsim14: forall (m: nat), fgsim 
     (Eoutput (S m);; (restart @ f) @ (S m ~- 1))
     (Eoutput (S m);; (restart @ g) @ (S m ~- 1))
| _fgsim15: forall (m: nat), fgsim 
     ((<>);; (restart @ f) @ (S m ~- 1))
     ((<>);; (restart @ g) @ (S m ~- 1))
| _fgsim16: forall (m: nat), fgsim 
     ((restart @ f) @ (S m ~- 1)) 
     ((restart @ g) @ (S m ~- 1))
.
Hint Constructors fgsim.

Lemma fg_simulated: 
  forall n m (EQ: n = 2 * m), similarity (f @ n) (g @ m).
Proof.
  i; eapply (@simul_tarski fgsim); [clear n m EQ|by eauto]. 
  i; destruct PR; subst; try by unfold_rfg; simul_step 1; fold_rfg.
  - by simul_step 0; fold_rfg; eauto.
  - by simul_step 0; fold_rfg; eauto.
  - by destruct m0; simul_step 2; eauto.
  - by destruct m; simul_step 1; eauto.
Qed.

